/****************************************************************************/
/*  Copyright 2005, Trustees of Indiana University                          */
/*  Author: Sidharth Thakur                                                 */
/*  Last modified: 14 January 2006                                          */
/*  License available at: <belt-trick-* folder>/docs/downloads/License.txt  */
/****************************************************************************/
#include "quaternion.h"
#include <stdio.h>

/* QUATERNION UTILITIES */

void quatBuildQuaternion ( double theta, double i, double j, double k, quaternion *q )
{
    q->r = cos(theta/2.0);
    q->i = i*sin(theta/2.0);
    q->j = j*sin(theta/2.0);
    q->k = k*sin(theta/2.0);
    //printf("qNorm = %f\n",qNorm(q));
}

void quatBuildQuaternion ( double theta, double* vector, quaternion *q )
{
    q->r = cos(theta/2.0);
    q->i = vector[0]*sin(theta/2.0);
    q->j = vector[1]*sin(theta/2.0);
    q->k = vector[2]*sin(theta/2.0);
    //printf("qNorm = %f\n",qNorm(q));
}

void quatGenerate ( quaternion *q, double theta, double *v)
{

	q->r = cos(theta/2.0);
	q->i = v[0]*sin(theta/2.0);
	q->j = v[1]*sin(theta/2.0);
	q->k = v[2]*sin(theta/2.0);
}

void quatMakeQuaternion     ( quaternion *q, double r, double i, double j, double k ) 
{
    q->r = r;
    q->i = i;
    q->j = j;
    q->k = k;
}

void quatMakeUnitQuaternion ( quaternion *q, double q0, double q1, double q2, double q3 ) 
{
    double l = sqrt( q0*q0 + q1*q1 + q2*q2 + q3*q3 );

    q->r = q0/l;
    q->i = q1/l;
    q->j = q2/l;
    q->k = q3/l;
}

void quatConjugate ( quaternion *q ) 
{
    q->r =  q->r;
    q->i = -q->i;
    q->j = -q->j;
    q->k = -q->k;
}

double	quatQuatDistance ( quaternion *q1, quaternion *q2 ) 
{

	return 2*acos( q1->r * q2->r  +  q1->i * q2->i  +  q1->j * q2->j  +  q1->k * q2->k );
}

double quatQuatDotProduct ( quaternion *q1, quaternion *q2 ) 
{
    return (q1->r*q2->r + q1->i*q2->i + q1->j*q2->j + q1->k*q2->k);
}

void quatQuatProduct ( quaternion *q1, quaternion *q2, quaternion *q3) 
{
    double w1,x1,y1,z1;
    double w2,x2,y2,z2;
    double w,x,y,z;

    // From Quaternion FAQ
    /* 
    w = w1w2 - x1x2 - y1y2 - z1z2
    x = w1x2 + x1w2 + y1z2 - z1y2
    y = w1y2 + y1w2 + z1x2 - x1z2
    z = w1z2 + z1w2 + x1y2 - y1x2 */

    w1 = q1->r;
    x1 = q1->i;
    y1 = q1->j;
    z1 = q1->k;
    
    w2 = q2->r;
    x2 = q2->i;
    y2 = q2->j;
    z2 = q2->k;
    
    w = w1*w2 - x1*x2 - y1*y2 - z1*z2;
    x = w1*x2 + x1*w2 + y1*z2 - z1*y2;
    y = w1*y2 + y1*w2 + z1*x2 - x1*z2;
    z = w1*z2 + z1*w2 + x1*y2 - y1*x2;

    q3->r = w;
    q3->i = x;
    q3->j = y;
    q3->k = z;
}


void quatToRotationMatrix ( quaternion *q, double *mat) 
{

    /*
              2     2                                      
       1 - (2Y  + 2Z )   2XY + 2ZW         2XZ - 2YW       
                                                           
                                2     2                    
    M= 2XY - 2ZW         1 - (2X  + 2Z )   2YZ + 2XW       
                                                           
                                                  2     2  
       2XZ + 2YW         2YZ - 2XW         1 - (2X  + 2Y ) 
                                                           
    */
	
    /*
     Q = |X Y Z W|
    */
    double X = q->i;
    double Y = q->j;
    double Z = q->k;
    double W = q->r;
    
	//printf("q2mQ = %f, %f, %f, %f\n",W,X,Y,Z);
    double xx, xy, xz, xw;
    double yy, yz, yw;
    double zz, zw;
    
    xx      = X * X;
    xy      = X * Y;
    xz      = X * Z;
    xw      = X * W;
    yy      = Y * Y;
    yz      = Y * Z;
    yw      = Y * W;
    zz      = Z * Z;
    zw      = Z * W;
    
    mat[0]  = 1 - 2 * ( yy + zz );
    mat[1]  =     2 * ( xy - zw );
    mat[2]  =     2 * ( xz + yw );
    mat[4]  =     2 * ( xy + zw );
    mat[5]  = 1 - 2 * ( xx + zz );
    mat[6]  =     2 * ( yz - xw );
    mat[8]  =     2 * ( xz - yw );
    mat[9]  =     2 * ( yz + xw );
    mat[10] = 1 - 2 * ( xx + yy );
    mat[3]  = mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
    mat[15] = 1;

    /*
      The resulting matrix uses the following positions:
      
       mat[0]  mat[4] mat[ 8] mat[12] 
    M= mat[1]  mat[5] mat[ 9] mat[13] 
       mat[2]  mat[6] mat[10] mat[14] 
       mat[3]  mat[7] mat[11] mat[15] 
    */
      //printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[0], mat[4], mat[8],  mat[12]);
      //printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[1], mat[5], mat[9],  mat[13]);
      //printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[2], mat[6], mat[10], mat[14]);
      //printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[3], mat[7], mat[11], mat[15]);
}

void	quatToRotationMatrixf	( quaternion *q, float *mat )
{

    /*
              2     2                                      
       1 - (2Y  + 2Z )   2XY + 2ZW         2XZ - 2YW       
                                                           
                                2     2                    
    M= 2XY - 2ZW         1 - (2X  + 2Z )   2YZ + 2XW       
                                                           
                                                  2     2  
       2XZ + 2YW         2YZ - 2XW         1 - (2X  + 2Y ) 
                                                           
    */
	
    /*
     Q = |X Y Z W|
    */
    float X = (float)q->i;
    float Y = (float)q->j;
    float Z = (float)q->k;
    float W = (float)q->r;
    
	//printf("q2mQ = %f, %f, %f, %f\n",W,X,Y,Z);
    float xx, xy, xz, xw;
    float yy, yz, yw;
    float zz, zw;
    
	//float a[3],b[3],c[3];

    xx      = X * X;
    xy      = X * Y;
    xz      = X * Z;
    xw      = X * W;
    yy      = Y * Y;
    yz      = Y * Z;
    yw      = Y * W;
    zz      = Z * Z;
    zw      = Z * W;
    
    mat[0]  = 1 - 2 * ( yy + zz );
    mat[1]  =     2 * ( xy - zw );
    mat[2]  =     2 * ( xz + yw );
    mat[4]  =     2 * ( xy + zw );
    mat[5]  = 1 - 2 * ( xx + zz );
    mat[6]  =     2 * ( yz - xw );
    mat[8]  =     2 * ( xz - yw );
    mat[9]  =     2 * ( yz + xw );
    mat[10] = 1 - 2 * ( xx + yy );
    mat[3]  = mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
    mat[15] = 1;

    /*
      The resulting matrix uses the following positions:
      
       mat[0]  mat[4] mat[ 8] mat[12] 
    M= mat[1]  mat[5] mat[ 9] mat[13] 
       mat[2]  mat[6] mat[10] mat[14] 
       mat[3]  mat[7] mat[11] mat[15] 
    */
	/*
	printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[0], mat[4], mat[8],  mat[12]);
	printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[1], mat[5], mat[9],  mat[13]);
	printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[2], mat[6], mat[10], mat[14]);
	printf("%.4f\t%.4f\t%.4f\t%.4f\n", mat[3], mat[7], mat[11], mat[15]);
	*/
	//transposeVec(mat);

	/*
	a[0] = mat[0]; a[1] = mat[1]; a[2] = mat[2]; 
	b[0] = mat[4]; b[1] = mat[5]; b[2] = mat[6]; 
	c[0] = mat[8]; c[1] = mat[9]; c[2] = mat[10];
	printf("dot(a,b) = %f\n",dot3D(a,b));
	printf("dot(a,c) = %f\n",dot3D(a,c));
	printf("dot(b,c) = %f\n",dot3D(b,c));
	*/
}

double qNorm ( quaternion *q ) 
{
    return ( sqrt(q->r*q->r + q->i*q->i + q->j*q->j + q->k*q->k ));
}

double qNorm2 (quaternion q) 
{
    return ( sqrt(q.r*q.r + q.i*q.i + q.j*q.j * q.k*q.k));
}

void quatFromRotationMatrix ( double *mat, quaternion *q ) 
{
    double T = 1 + mat[0] + mat[5] + mat[10]; //Matrix Trace
    double S,W,X,Y,Z;

    if(T > 0.00000001) {
        S = sqrt(T) * 2;
        W = 0.25 * S;
        X = ( mat[9] - mat[6] ) / S;
        Y = ( mat[2] - mat[8] ) / S;
        Z = ( mat[4] - mat[1] ) / S;
    }
    else {

        if ( mat[0] > mat[5] && mat[0] > mat[10] )  {	// Column 0: 
            S  = sqrt( 1.0 + mat[0] - mat[5] - mat[10] ) * 2;
            X = 0.25 * S;
            Y = (mat[4] + mat[1] ) / S;
            Z = (mat[2] + mat[8] ) / S;
            W = (mat[9] - mat[6] ) / S;
        } else if ( mat[5] > mat[10] ) {			// Column 1: 
            S  = sqrt( 1.0 + mat[5] - mat[0] - mat[10] ) * 2;
            X = (mat[4] + mat[1] ) / S;
            Y = 0.25 * S;
            Z = (mat[9] + mat[6] ) / S;
            W = (mat[2] - mat[8] ) / S;
        } else {						// Column 2:
            S  = sqrt( 1.0 + mat[10] - mat[0] - mat[5] ) * 2;
            X = (mat[2] + mat[8] ) / S;
            Y = (mat[9] + mat[6] ) / S;
            Z = 0.25 * S;
            W = (mat[4] - mat[1] ) / S;
        }
        
    }
    q->r = W;
    q->i = X;
    q->j = Y;
    q->k = Z;
}

//////////////////////////////////////////////////////////////
// Rotation Matrix to Quaternions
//////////////////////////////////////////////////////////////
void ExtractQuaternionFromMat ( double m[16], quaternion *q )
{
	double tr = 0.0;
	double q0,q1,q2,q3;
	
	tr = 1 + m[0] + m[5] + m[10];

	if(tr > 0.000000001) {
		q0 = sqrt(tr)/2.0;
		q1 = (m[6] - m[9])/(4.0*q0);
		q2 = (m[8] - m[2])/(4.0*q0);
		q3 = (m[1] - m[4])/(4.0*q0);
	}
	else {
		q1 = 0.5*sqrt( 1 + m[0] - m[5] - m[10] );
		q2 = 0.5*sqrt( 1 + m[5] - m[10] - m[0] );
		q3 = 0.5*sqrt( 1 + m[10] - m[0] - m[5] );

		if(q1>= q2 && q1>= q3) {
			q0 = (m[6] - m[9])/(4.0*q1);
			q2 = (m[1] + m[4])/(4.0*q1);
			q3 = (m[8] + m[2])/(4.0*q1);
		}
		else if(q2>= q3 && q2 >= q1) {
			q0 = (m[8] - m[2])/(4.0*q2);
			q1 = (m[4] + m[1])/(4.0*q2);
			q3 = (m[9] + m[6])/(4.0*q2);
		}
		else if(q3>= q1 && q3>= q2) {
			q0 = (m[1] - m[4])/(4.0*q3);
			q1 = (m[8] + m[2])/(4.0*q3);
			q2 = (m[6] + m[9])/(4.0*q3);
		}
	}
	q->r=q0;
	q->i=q1;
	q->j=q2;
	q->k=q3;
}

void quatNormalize ( quaternion *q ) 
{
    double len = qNorm(q);
    q->r /= len;
    q->i /= len;
    q->j /= len;
    q->k /= len;
}

void quatCopy ( const quaternion *q1, quaternion *q2 ) 
{
    q2->r = q1->r;
    q2->i = q1->i;
    q2->j = q1->j;
    q2->k = q1->k;
}

//////////////////////////////// End Of File //////////////////////////////////////////////
